#!/usr/bin/perl
#_*****************************************************************************
#_*                                                                           *
#_* File name: makeCpGFig3.cgi                                                *
#_* cgi for creating methylation status graph 3                               *
#_* Author: Yuichi Kumaki (kuma@cdb.riken.go.jp or yuichi@kumaki.jp)          *
#_* Copyright (C) 2008 Yuichi Kumaki                                          *
#_* Laboratory for Mammalian Epigenetic Studies,                              *
#_* Center for Developmental Biology, RIKEN                                   *
#_* 2-2-3 Minatojima-Minamimachi, Chuo-ku, Kobe 650-0047, Japan               *
#_*                                                                           *
#_* 2008/02/20 First open source version                                      *
#_*                                                                           *
#_*****************************************************************************
#_*                                                                           *
#_* This program is free software: you can redistribute it and/or modify      *
#_* it under the terms of the GNU General Public License as published by      *
#_* the Free Software Foundation, either version 3 of the License, or         *
#_* (at your option) any later version.                                       *
#_*                                                                           *
#_* This program is distributed in the hope that it will be useful,           *
#_* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
#_* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
#_* GNU General Public License for more details.                              *
#_*                                                                           *
#_* You should have received a copy of the GNU General Public License         *
#_* along with this program.  If not, see <http://www.gnu.org/licenses/>.     *
#_*                                                                           *
#_*****************************************************************************

use 5.6.0;
use strict;
use warnings;

use GD;
use CGI::Lite;
use BFSetting;
use BFUtils;

my ($cgi  );
my ($data );
my ($ip   );
my (@pos  );
my (@val  );
my ($i    );
my (%meth );
my ($gwid );
my ($width);
my ($hight);
my ($gd   );
my ($white);
my ($black);
my ($gray );
my ($mag  );
my ($x    );
my ($x1   );
my ($x2   );
my ($j    );
my ($name );
my ($fig  );
my ($size );
local ($_ );

$cgi = new CGI::Lite;
$data = $cgi->parse_form_data;
$ip = join '', map {sprintf "%02X", $_} split /\./, $ENV{REMOTE_ADDR};
$data->{id} =~ /^\d{20}$ip$/
    or errorAndExit ('Invalid access', 'Invalid access: IP address is different from your data submission access, or your browser cache data is expired. Please resubmit sequence!', "Invalid ID or different IP : $data->{id} : $ip"); 

$data->{proj} ||= '';
$data->{proj} =~ y/&;`'"|*?~<>^()[]{}$\\\n\r//d;
$data->{proj} =~ s/[^\w-]//g;

@pos = split ',', $data->{pos};
@val = split ',', $data->{val};

@pos == @val
    or errorAndExit ('Parameter error!', 'Parameter error!', "Parameter error! : \@pos != \@val : " . @pos . " != " . @val);

for ($i = 0; $i < @pos; $i++) {
    $pos[$i] =~ /^\d+$/
	or errorAndExit ('Parameter error!', 'Parameter error!', "Parameter error! : \$pos[$i] = $pos[$i]");
    $meth{$pos[$i]} = {};
    ($meth{$pos[$i]}->{meth}, $meth{$pos[$i]}->{total}) = split '\|', $val[$i];
    $meth{$pos[$i]}->{meth} =~ /^\d+$/
	or errorAndExit ('Parameter error!', 'Parameter error!', "Parameter error! : \$meth{$pos[$i]}->{meth} = $meth{$pos[$i]}->{meth}");
    $meth{$pos[$i]}->{total} =~ /^\d+$/
	or errorAndExit ('Parameter error!', 'Parameter error!', "Parameter error! : \$meth{$pos[$i]}->{total} = $meth{$pos[$i]}->{total}");
    $meth{$pos[$i]}->{total} ||= 1;
    $meth{$pos[$i]}->{ratio} = $meth{$pos[$i]}->{meth} / ($meth{$pos[$i]}->{total});
}

$data->{len} =~ /^\d+$/
    or errorAndExit ('Parameter error!', 'Parameter error!', "Parameter error! : \$data->{len} = $data->{len}");

$gwid  = ($data->{len} - @pos ) * 2 + @pos * GWIDTH;
$width = $gwid + MARGL + MARGR;

$hight = HIGHT;
$gd = new GD::Image ($width, $hight);
$white = $gd->colorAllocate (255,255,255);
$black = $gd->colorAllocate (  0,  0,  0);
$gray  = $gd->colorAllocate (128,128,128);
$mag   = $gd->colorAllocate (255,  0,255);

$gd->rectangle (MARGL - 1, GMARGH,
		MARGL + $gwid,  HIGHT - GMARGB2,
		$black);

$gd->string   (gdSmallFont,
	       LFMARG, HIGHT - BMARG - FONTH2,
	       'CpG position', $black);
$gd->string   (gdSmallFont,
	       MARGL - LLABLEN - 1, GMARGH - FONTH2,
	       '100%', $black);
$gd->string   (gdSmallFont,
	       MARGL - LLABLEN - 1, HIGHT - GMARGB2 - FONTH2,
	       '  0%', $black);
$gd->stringUp (gdSmallFont,
	       MARGL - LLABLEN - LLABMRG - 1, HIGHT - GMARGB2 - FONTH2 - 3,
	       'Methylated (%)', $black);

for ($i = 0; $i < @pos; $i++) {

    $x = MARGL + ($pos[$i] - $i - 1) * 2 + GWIDTH * $i;
    $x1 = $x;
    $x2 = $x1 + GWIDTH - 1;

    $gd->stringUp (gdTinyFont,
		   $x1 + 1, HIGHT - 7,
		   $pos[$i], $black);

    if ($meth{$pos[$i]}->{ratio}) {
	for ($j = $x1; $j <= $x2; $j++) {
	    $gd->line ($j,
		       HIGHT - GMARGB2 - 1,
		       $j,
		       int (HIGHT - GMARGB2 - (HIGHT - GMARGB2 - GMARGH) * $meth{$pos[$i]}->{ratio}),
		       $gray);
	}
    } else {
	$gd->line ($x1,
		   HIGHT - GMARGB2,
		   $x2,
		   HIGHT - GMARGB2,
		   $mag);
    }
}

$fig = $gd->png;
$size = length $fig;

if ($data->{down}) {
    $name = substr $data->{id}, 0, 12;
    $name .= "_$data->{proj}" if ($data->{proj});
    $name .= "_CpG_status_graph3.png";

    print qq{Content-Disposition: attachment; filename="$name"\n};
    print qq{Content-Length: $size\n};
}

print qq{Content-Type: image/png\n};
print qq{Status: 200 OK\n\n};
print $fig;

###########################  End of Main ######################################
exit; #--1----+----2----+----3----+----4----+----5----+----6----+----7----+----
###########################  End of Main ######################################

#---+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----
###############################################################################
1;
###############################################################################
